"""
This module defines an UFL transformation, that takes a UFL expression
and transforms it into a sum of accumulation terms.
"""
from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.ufl.transformations import ufl_transformation
from dune.perftool.ufl.transformations.replace import replace_expression
from dune.perftool.ufl.transformations.identitypropagation import identity_propagation
from dune.perftool.ufl.transformations.reindexing import reindexing
from dune.perftool.ufl.modified_terminals import ModifiedArgumentDescriptor

from ufl.classes import Zero, Identity, Indexed, IntValue, MultiIndex
from ufl.core.multiindex import indices

from pytools import Record


class AccumulationTerm(Record):
    def __init__(self, term, argument):
        Record.__init__(self, term=term, argument=argument)


@ufl_transformation(name="accterms", extraction_lambda=lambda l: [at.term for at in l])
def split_into_accumulation_terms(expr):
    ret = []

    # Extract a list of modified terminals for the test function
    # One accumulation instruction will be generated for each of these.
    test_args = extract_modified_arguments(expr, argnumber=0)

    # Extract a list of modified terminals for the ansatz function
    # in jacobian forms. Only the restriction of those terminals will
    # be used to generate new accumulation terms!
    all_jacobian_args = extract_modified_arguments(expr, argnumber=1)

    for test_arg in test_args:
        # Do this as a multi step replacement procedure to avoid UFL nagging
        # about too much stuff in reconstructing the expressions

        # 1) We first cut the expression to the relevant modified test_function
        # Build a replacement dictionary
        replacement = {ma.expr: Zero() for ma in test_args}
        replacement[test_arg.expr] = test_arg.expr
        replace_expr = replace_expression(expr, replacemap=replacement)

        # 2) Cut the test function itself from the expression
        if test_arg.reference_grad:
            newi = indices(1)
            replacement = {test_arg.expr: Indexed(Identity(2), MultiIndex(newi + test_arg.index._indices))}
            test_arg = ModifiedArgumentDescriptor(reindexing(test_arg.expr, replacemap={test_arg.index[0]: newi[0]}))
        else:
            replacement = {test_arg.expr: IntValue(1)}
        replace_expr = replace_expression(replace_expr, replacemap=replacement)

        # 3) Collapse any identity nodes that may have been introduced by replacing vectors
        replace_expr = identity_propagation(replace_expr)

        # 4) Further split according to trial function in jacobian terms
        if all_jacobian_args:
            for jac_arg in all_jacobian_args:
                pass
        else:
            if not isinstance(replace_expr, Zero):
                ret.append(AccumulationTerm(replace_expr, test_arg))

    return ret
